/* 
In this exercise you will walk through a scenario where a relatively low percentage
of rows are modified and the fragmentation created as a result is high.
*/
USE Credit
go

--1) Create a working version of the member table called member2. 

IF OBJECTPROPERTY(object_id('member2'), 'IsUserTable') = 1
	DROP TABLE dbo.member2 
go
CREATE TABLE dbo.member2 (
	member_no 	numeric_id 		NOT NULL ,
	lastname 		shortstring 	NOT NULL ,
	firstname 		shortstring 	NOT NULL ,
	middleinitial 	letter 			NULL ,
	street 			shortstring 	NOT NULL ,
	city 			shortstring 	NOT NULL ,
	state_prov 		statecode 		NOT NULL ,
	country 		countrycode 	NOT NULL ,
	mail_code 		mailcode 		NOT NULL ,
	phone_no 		phonenumber 	NULL ,
	issue_dt 		datetime 		NOT NULL ,
	expr_dt 		datetime 		NOT NULL ,
	region_no 		numeric_id 		NOT NULL ,
	corp_no 		numeric_id 		NULL ,
	prev_balance 	money 			NULL ,
	curr_balance 	money 			NULL ,
	member_code 	status_code 	NOT NULL 
)
GO

--2) Populate the member2 table with the same data from member.

INSERT member2
	SELECT * FROM member
go

--3) Create indexes on the member2 table to simulate a realworld environment.

ALTER TABLE member2
ADD CONSTRAINT member2PK
	PRIMARY KEY CLUSTERED (member_no)
go

CREATE INDEX member2NameInd 
ON member2(lastname, firstname, middleinitial)
go

CREATE INDEX member2RegionFK
ON member2(region_no)
go

CREATE INDEX member2CorpFK
ON member2(corp_no)
go

-- Verify the indexes
sp_helpindex member2
go
select object_name(id) as tablename, indid, name
from sysindexes
where object_id('member2') = id

--4) Verify the fragmentation of the indexes

-- Prior to SQL Server 2000
DECLARE @ObjID	int
SELECT @ObjID = object_id('member2')
DBCC SHOWCONTIG(@ObjID)
--or
-- New Syntax in SQL Server 2000
DBCC SHOWCONTIG('member2') --WITH TABLERESULTS (new FINALLY in ss2k)

	/* NOTE: there should not be any fragmentation currently as the indexes 
	have just been built.
        */

-- 5) Simulate modifications against the member2 table by performing a single update
-- statement that updates roughly 5% of the table. Also note that the script shows the 
-- total time it takes to perform the modification (just as an fyi).

DECLARE @StartTime	datetime, 
	@EndTime	datetime
SELECT @StartTime = getdate()
UPDATE member2
	SET street = '1234 Main St.',
		city = 'Anywhere'
WHERE member_no % 19 = 0
SELECT NumRowsMod = @@RowCount
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
go

-- 6) Next let's look at the [negative] affect that this has had on our member2 table.

DBCC SHOWCONTIG('member2') --WITH TABLERESULTS (new FINALLY in ss2k)
go
-- 	NOTE: the fragmentation should be extreme as the Scan Density will probably 
-- 	show a VERY low number. 

-- 7) Should we rebuild our index? And should we set a FILLFACTOR? 

-- 	ABSOLUTELY - YES and YES! 

-- 8) So how do we choose our fillfactor?

-- Lecture?

-- 9) Now that we've decided it is important to rebuild our indexes we need to look at 
-- options. There are multiple ways to rebuild an index, all of which are shown below with
-- syntax and notes on why you might use one over the other. For this lab we are going to 
-- use the most common option - CREATE WITH DROP_EXISTING.

-- 	* DROP and RECREATE
-- 	    NOTES - This is more challenging than it sounds considering our clustered index is our Primary Key. 
-- 		We cannot drop this Primary Key IF there are foreign keys which reference it unless we drop all of the
-- 		foreign keys first. Since this is NOT realistic SQL Server has DBCC DBREINDEX and CREATE with 
-- 		DROP_EXISTING. Additionally because the non-clustered indexes depend on the clustered 
--		index value dropping and recreating will cause the non-clustered indexes to be rebuilt TWICE!
-- 
-- 	* DBCC DBREINDEX (added in SQL Server 6.5)
-- 	    NOTES - This is in the product primarily for backward compatibility however it DID have one big difference
-- 		over CREATE with DROP_EXISTING in that it will force ALL non-clustered indexes to be rebuilt with
-- 		the specified fillfactor. This is sometimes desirable and sometimes not. Often the non-clustered indexes
-- 		will gain from having a fillfactor but probably NOT the same fillfactor as the clustered index. Remember
-- 		a CL index IS the data and in fact most changes occur here. The non-clustered indexes only contain
-- 		part of the data (the nonclustered index key columns plus the clustered key columns). Sometimes the 
-- 		NC indexes have more fragmentation sometimes less. It depends on the columns in the index and 
--		whether or not those columns are volitale. The best way to run DBCC DBREINDEX is to run it without
-- 		a fillfactor and in fact all indexes get rebuilt with their Original FIll Factor. This "OrigFillFactor" is the 
--		same fillfactor that was used the last time the index was built. This gives you the greatest flexibility.

--		NOTE: As of 2000sp2 - the nonclustered indexes are NOT rebuilt using DBCC DBREINDEX. In 2000sp2
--		DBCC DBREINDEX becomes almost identical to CREATE with DROP_EXISTING EXCEPT that it's easier
--		to automate.
--
--		SO - my recommendation is to use DBCC DBREINDEX because of it's simplicity to automate!
-- 
-- 	    syntax
		DBCC DBREINDEX (member2, member2PK, 85) 
-- 		DBCC DBREINDEX (member2, member2PK) -- Use the OrigFillFactor
--
-- 	* CREATE with DROP_EXISTING (new in SQL Server 7.0)
-- 	    NOTES - This was added to minimize the recreation of non-clustered indexes when only the leaf level of the
-- 		clustered was fragmented. In 2000sp2 - this is virtually the same as DBCC DBREINDEX. Pre-sp2 it was a bug
--		that all of the non-clustered indexes were rebuilt.

--        CREATE WITH DROP_EXISTING when you DO NOT CHANGE the definition of the CL index
--            WILL NOT rebuild the nonclustered indexes UNLESS you are rebuilding the CL index on a non-unique key
--        CREATE WITH DROP_EXISTING when you CHANGE the definition of the CL index
--            WILL rebuild the nonclustered indexes using their OrigFillFactor (from sysindexes)

-- 	    syntax
-- 		CREATE UNIQUE CLUSTERED INDEX member2pk
-- 			ON member2(member_no)
-- 			WITH FILLFACTOR = 85, DROP_EXISTING
-- 
--	IF original index is 
-- 		CREATE  CLUSTERED INDEX member2CL
-- 			ON member2(lastname, firstname, middleinitial)

-- 		CREATE UNIQUE CLUSTERED INDEX member2CL  --requires the same name
-- 			ON member2(member_no)
-- 			WITH FILLFACTOR = 85, DROP_EXISTING
--
-- 	The best time to perform ANY of these modifications is when the system is at it lowest usage. All of the above options
--	place an EXCLUSIVE Lock on the base table (unless you're building a nonclustered index and then in that case it's a 
--	SHARED lock on the base table). BUT wait...there's a 2000 option!

--	If the system is a 24x7 system then SQL Server 2000 offers a new option called DBCC INDEXDEFRAG. 
--
-- 	* DBCC INDEXDEFRAG (new in SQL Server 2000)
-- 	    NOTES - This option, when a table is heavily fragmented, will take longer to defrag a table than the CREATE
-- 		with DROP_EXISTING however it will allow access to the table while it is being rebuilt. This option relies on an 
--		already set FillFactor...so you must rebuild the index once OR you must have created the index with the fillfactor set.
--		CREATE TABLE does support FILLFACTOR settings when creating the clustered and nonclustered indexes through
--		constraints. TRICK - You can also modify the system table to make this change---change the sysindexes table for the
--		OrigFillFactor column. Make sure you verify your SQL update by writing it as a select first!
-- 
--   	    syntax
--            DBCC INDEXDEFRAG(credit, member2, member2cl)


-- 10) Once recreated, check the fragmentation again.

DBCC SHOWCONTIG('member2')

-- 	NOTE: the fragmentation is what we expect - 90%. As that's what we chose for 
-- 	our fillfactor.

-- 11) Let's try the modification again with even more rows.

DECLARE @StartTime	datetime, 
	@EndTime	datetime
SELECT @StartTime = getdate()
UPDATE member2
	SET street = '1234 Main St.',
		city = 'Anywhere'
WHERE member_no % 17 = 0
SELECT NumRowsMod = @@RowCount
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
go

--12) And again, check the fragmentation to see if our chosen fillfactor was appropriate.

DBCC SHOWCONTIG('member2')

--	NOTE: And yes, this modification did not create fragmentation.
-- 		You can keep making modifications and see how many you get before you begin fragmenting your data again!

-- 13) Another way to see information about indexes from sysindexes

SELECT CASE
		WHEN indid = 0 THEN 'Heap'
		WHEN indid = 1 THEN 'CL'
		ELSE 'NC ' + convert(varchar(5), indid)
	END AS IndexID, 
	object_name(id) AS ObjectName,
	name AS IndexName,
	OrigFillFactor AS 'FillFactor'
FROM sysindexes
    WHERE indid <= 250
ORDER BY 2,1

-- HAVE FUN!!!!!
-- Kimberly